/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "fmr.h"
#include "fmr_mem.h"
#include "fmr_cost.h"
#include "fmr_tools.h"
#include "fmr_exploRandSwap.h"
#include "fmr_exploShift.h"
#include "fmr_exploSwap.h"
#include "fmr_exploEjection.h"
#include "fmr_vns.h"
#include "fmr_recuit.h"
#include "fmr_recuit_b.h"
#include "fmr_recuit_c.h"
#include "fmr_glouton.h"
#include "fmr_core.h"
#include "fmr_threads.h"
#include "fmr_profondeur.h"
#include "fmr_relax.h"
#include "fmr_tirage.h"

/**
 * Le calcul lui-meme.
 * Prend en entree un pointeur vers l'instance, un pointeur vers
 * le placement initial et un pointeur vers le resultat du calcul.
 * Retourne FMR_OK si tout s'est bien passe, une erreur si il 
 * y a eu un probleme.
 */

int
calcul_assignement (const placement_t * initial, const placement_t * initial_relax, const char *listeAlgos, unsigned int seed) {
    int64 gain=0;
    unsigned int i ;
    long gainMin = 0;		/* a parametrer dynamiquement */
    unsigned short longChaine = MAXLONGCHAINEEJECTION;	/* choisi par experimentation */
    placement_t resultat_courant;
    struct drand48_data rand_buf ;
    tirage_t tirage;		// tirage (aléatoire ou pas) d'un mouvement
    unsigned max_essais = 10000;

    init_rand(seed, &rand_buf);
    
    resultat_courant = alloc_assignement(instance.nb_procs);
    
    copier_assignement(initial, &resultat_courant);
    
    for (i = 0; *listeAlgos!=0; i++,listeAlgos++) {
        switch (*listeAlgos) {
	    case '.':
		copier_assignement(initial,  &resultat_courant );
		gain = 0;
		break;
            case 'I':
                gain = calcul_assignement_iter(initial, &resultat_courant, gainMin);
                break;
            case 'R':
                gain = explo_voisinageRand (initial, &resultat_courant, gainMin, &rand_buf);	//Random
                break;
            case 'E':
                DEBUG_PRINT ("Explo%d: swap (%ld)\n", i, gainMin);
                gain = explo_voisinageSwap (initial, &resultat_courant, gainMin);	//Echange
                break;
            case 'X':
                DEBUG_PRINT ("Explo%d: VeryBestswap (%ld)\n", i, gainMin);
                gain = explo_voisinageVeryBestSwap(initial, &resultat_courant, gainMin);	//Echange
                break;
            case 'W':
                DEBUG_PRINT ("Explo%d: Bestswap (%ld)\n", i, gainMin);
                gain = explo_voisinageBestSwap(initial, &resultat_courant, gainMin);	//Echange
                break;
            case 'S':
                DEBUG_PRINT ("Explo%d: Shift (%ld)\n", i, gainMin);	// Shift
                gain = explo_voisinageShift (initial, &resultat_courant, gainMin);
                break;
            case 'B':
                DEBUG_PRINT ("Explo%d: BestShift (%ld)\n", i, gainMin);
                gain = explo_voisinageBestShift (initial, &resultat_courant, gainMin);	//Best shift
                break;
            case 'Y':
                DEBUG_PRINT ("Explo%d: VeryBestShift (%ld)\n", i, gainMin);
                gain = explo_voisinageVeryBestShift (initial, &resultat_courant, gainMin);// Very Best shift
                break;
            case 'C':
                DEBUG_PRINT ("Explo%d: EjectionChain (%ld)\n", i, gainMin);
                gain = explo_voisinageEjectionChain (initial, &resultat_courant, longChaine,
                                                   gainMin);
                break;
            case 'A':
                DEBUG_PRINT ("Explo%d: Recuit (%ld)\n", i, gainMin);
                gain = explo_recuit_c (initial, &resultat_courant, &rand_buf);
                break;
            case 'L':
                DEBUG_PRINT ("Explo%d: Recuit Best Loop\n", i);
                gain = explo_recuit_best_loop_b (initial, &rand_buf);
                break;
            case 'N':
                DEBUG_PRINT ("Explo%d: Recuit Best Loop\n", i);
                gain = explo_recuit_best_loop_c (initial, &rand_buf);
                break;
            case 'G':
		if (instance.nb_transcient>0) break;
                DEBUG_PRINT ("Explo%d: Glouton\n", i);
                //gain = explo_glouton (initial, &resultat_courant);
		gain=explo_GloutonReduction(initial, &resultat_courant, gainMin);
                break;
            case 'V':
                DEBUG_PRINT ("Explo%d: VNS\n", i);
                gain = calcul_assignement_vns (initial, initial_relax, &resultat_courant, &seed);
                break;
            case 'P':
                DEBUG_PRINT ("Explo%d: Prof.\n", i);
                gain = calcul_assignement_rec (initial, &resultat_courant);
                break;
            case 'Q':
                DEBUG_PRINT ("Explo%d: Prof. Max \n", i);
                gain = calcul_assignement_rec_max (initial, &resultat_courant, *(++listeAlgos)-'0');
                break;
            case 'M':
                DEBUG_PRINT ("Explo%d: SC Relax \n", i);
                gain = calcul_assignement_relax (initial_relax, &resultat_courant, &seed);
                break;
            case 'D':
                DEBUG_PRINT ("Explo%d: Remove Add Procs Glouton\n", i);
                gain = explo_rap_glouton (initial, &resultat_courant);
                DEBUG_PRINT ("Remove Add Procs Glouton Gain=%lld\n", gain);
                break;
            case 'F': // F
                DEBUG_PRINT ("Explo%d: Shift Elementaire\n", i);
		tirage = tirage_shift_checked(&resultat_courant, max_essais, &rand_buf);
		if (tirage.nouveau)	// transition valide trouve
		{
		  gain = calcule_gain_tirage (initial, &resultat_courant, tirage);
		  if (gain>0)
		  {
		    DEBUG_PRINT ("Shift Elementaire ok, gain=%ld\n", gain);
		    effectue_tirage (initial, &resultat_courant, tirage);		//Shift elementaire
		    resultat_courant.cout -= gain ;
		  }
		}
                break;
            case 'H': // H
                DEBUG_PRINT ("Explo%d: Swap Elementaire\n", i);
		tirage = tirage_swap_checked(&resultat_courant, max_essais, &rand_buf);
		if (tirage.nouveau)	// transition valide trouve
		{
		  gain = calcule_gain_tirage (initial, &resultat_courant, tirage);
		  if (gain>0)
		  {
		    DEBUG_PRINT ("Swap Elementaire ok, gain=%ld\n", gain);
		    effectue_tirage (initial, &resultat_courant, tirage);		//Swap elementaire
		    resultat_courant.cout -= gain ;
		  }
		}
                break;
	    case 'Z':
		DEBUG_PRINT ("Explo%d: GuidedShift\n",i);
		gain = explo_voisinageShift2(initial, &resultat_courant, gainMin);
		break;
            default:
                fprintf (stderr, "Erreur algo inconnu\n");
		gain = 0;
                break;
        }
        if (gain>gainMin && *listeAlgos!='A' && *listeAlgos!='V' && *listeAlgos!='Q' && (*listeAlgos<'0' || *listeAlgos>'9'))
	    atomic_set_meilleur_placement(&resultat_courant, *listeAlgos);
    }
    if (alarme)
        return FMR_TIMEOUT;
    
    free_placement( &resultat_courant );
    return (int) gain;
}

#define NBITER 20

long
calcul_assignement_iter (const placement_t * initial, placement_t * resultat, long gainMin) {
    int j;
    long gainMax, res;		/* a parametrer dynamiquement */
    unsigned short longChaine = MAXLONGCHAINEEJECTION;
    placement_t tempo = alloc_assignement( instance.nb_procs );
    placement_t tempoMax = alloc_assignement( instance.nb_procs );
    char meilleurAlgo;
    
    for (j=0; j<NBITER; j++) {
        //res = explo_voisinageRand (initial, resultat, gainMin);	//Random
        copier_assignement(resultat, &tempo);
        gainMax = explo_voisinageSwap (initial, &tempo, gainMin);	//Echange
        meilleurAlgo='E';
        
        copier_assignement(&tempo, &tempoMax);
        copier_assignement(resultat,&tempo);
        if (gainMax < (res = explo_voisinageEjectionChain (initial, &tempo, longChaine, gainMin))) {
            gainMax=res;
            copier_assignement(&tempo, &tempoMax);
            meilleurAlgo='C';
        }
        
        copier_assignement(resultat,&tempo);
        if (gainMax < (res = explo_voisinageShift (initial, &tempo, gainMin))) {
            gainMax=res;
            copier_assignement(&tempo, &tempoMax);
            meilleurAlgo='S';
        }
        
        copier_assignement(resultat,&tempo);
        if (gainMax < (res = explo_voisinageBestShift (initial, &tempo, gainMin))) {
            gainMax=res;
            copier_assignement(&tempo, &tempoMax);
            meilleurAlgo='B';
        }
        
        copier_assignement(resultat,&tempo);
        if (gainMax < (res = explo_voisinageBestSwap(initial, &tempo, gainMin))) {
            gainMax=res;
            copier_assignement(&tempo, resultat); // c'est bien resultat ici 
            meilleurAlgo='W';
        } else {
            copier_assignement(&tempoMax, resultat);
        }
        
        DEBUG_PRINT("Meilleur algo:%c, gain=%ld\n",meilleurAlgo,gainMax);
        DEBUG_PRINT("%c:%ld\n",meilleurAlgo,gainMax);
        if (gainMax==0) {
            DEBUG_PRINT("nbre d'iteration:%d",j);
            break;
        }
        
    }
    return gainMax;
}
