/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include "fmr.h"
#include "fmr_exploRandSwap.h"
#include "fmr_tools.h"
#include "fmr_mem.h"
#include "fmr_cost.h"
#include "fmr_constraints.h"
#include "fmr_psmvt.h"

#include <pthread.h>

/* prototypes des fonctions internes */
unsigned long  cout_deplacements( unsigned short* );
bool une_machine_est_surchargee(placement_t*);
unsigned int machine_surchargee( placement_t* );

/* Constantes utilisees ici */
//#define VOISINAGEMAX     5
#define VOISINAGEMAX     5u
//#define NBITER        1000
#define NBITER           3000u

/**
 * Algo qui explore un voisinage
 * On bouge un processus de cout minimal, et on fait ca NBITER fois.
 * Si le cout est inferieur au placement precedent, on bouge.
 */

long explo_voisinageRand( const placement_t const * initial, placement_t* resultat, long gainMin, struct drand48_data * rand_buf ) {

    placement_t pl_temp;
    int64 cout_init;//, new_cout;
    unsigned int min_cout_depl;
    unsigned int i, j, k, m;
    unsigned short proc_rank, origine, destination, proc_rank_binome;

	long gain, gainTotal=0;
    unsigned short deplacements[VOISINAGEMAX];

    pl_temp = alloc_assignement( instance.nb_procs );

    copier_assignement( resultat, &pl_temp );

    /* Chercher le cout de deplacement mini d'un processus
       On ne tient pas compte des couts de migration vers les voisins */

    min_cout_depl = 0x0FFFFFFF;
    for( i = 0 ; i < instance.nb_procs ; i++ ) {
        min_cout_depl = MIN( min_cout_depl, instance.procs[i].PMC );
    }


    /* on ameliore la solution acceptable */

    for( j = 1 ; j <= VOISINAGEMAX ; j++ ) {
	DEBUG_PRINT("voisinage=%d\n",j);

        /* En bouclant ici, on agrandit la taille du voisinage
           en jouant sur le cout de deplacement d'un processus :
           min_cout_depl puis min_cout_depl *2 puis min_cout_depl*3
           etc, jusqu'a min_cout_depl * VOISINAGEMAX.
        */

        for( i = 0 ; i < NBITER ; i++ ) {

            cout_init = initial->cout;

            /* On en tire un aleatoirement un ensemble de procs a
               deplacer jusqu'a ce que la somme de leurs couts
               depasse le cout max acceptable. On deplace au maximum
               VOISINAGEMAX processus.
            */

            for( k = 0 ; k < VOISINAGEMAX ; k++ ) {
                deplacements[k] = NOPROCESS;
            }

            for( k = 0 ; k < VOISINAGEMAX ; k++ ) {
            tirage_processus:


                /* On commence par deplacer les procs des machines surchargees */

                deplacements[k] = (unsigned short) irand_a_b( rand_buf, 0, instance.nb_procs );

                //DEBUG_PRINT( "on a tire %d,", deplacements[k] );
                if( cout_deplacements( deplacements ) > j * min_cout_depl ){
                    if( k > 0 ) {
                        deplacements[k] = NOPROCESS;
                        k--;
                        break;
                    } else {
                        /* si on ne fait pas ca on se retrouve avec plein de passages avec 0 deplacements */
                        goto tirage_processus;
                    }
                }
            }
            //DEBUG_PRINT( "on a tire %d deplacements\n", k+1 );

            for( m = 0 ; m < k ; m++ ) {
		/*pthread_testcancel();*/

                if( deplacements[m] != NOPROCESS ) {

#ifdef DEBUG
                    machine_t* machdest;
                    machine_t* machorig;
#endif

                    proc_rank = deplacements[m];
                    origine = pl_temp.assignment[ proc_rank ];

                    /* On retire tant qu'on tombe sur la meme machine : on veut bouger */
                    do{
                        destination = (unsigned short) irand_a_b( rand_buf, 0, instance.nb_machines );
                    } while( destination == origine );

#ifdef DEBUG
                    machdest = &( instance.machines[ destination ]);
                    machorig = &( instance.machines[ pl_temp.assignment[ proc_rank ] ] );
#endif

#if 0
                    /* Est-ce qu'il reste de la capacite pour accueillir ce proc sur la machine ? */

                    rc = ajouter_proc_old( instance, &pl_temp, destination, proc_rank, true );
                    /*                 if( FMR_OVERLOAD == rc ) {
                        DEBUG_PRINT( "La machine %d va etre surchargee si on bouge - on ne bouge pas le processus\n", destination );
                        }*/

                    if( FMR_OK == rc ) {
                        rc = retirer_proc( instance, origine, proc_rank );
                        DEBUG_PRINT( "on bouge le processus %d de la machine %d vers la machine %d\n", proc_rank, origine, destination );
                        DEBUG_PRINT( "la machine %d (destination) est maintenant chargee a %d/%d en RAM et %d/%d en CPU ; la machine %d (origine) est maintenant chargee a %d/%d en RAM et %d/%d en CPU\n", destination, machdest->U[1], machdest->capa[1], machdest->U[0], machdest->capa[0], resultat->assignment[ proc_rank ], machorig->U[1], machorig->capa[1], machorig->U[0], machorig->capa[0] );
                    }
#endif

                    /* on tire aleatoirement un proc sur la machine destination */

                    //proc_rank_binome = rand_intervalle( 0, instance.machines[destination].nb_procs - 1 );
		    // ARG si on fait cela il faut passer en revue le liste de procs de la machine destination pour trouver
		    // le numero de proc correspondant: lourd; pour l'instant prenons le premier.
		    while(listeCompacte_est_vide(& pl_temp.liste_procs,destination)) {// he oui cela peut arriver :-(
			destination ++;
			if (destination > instance.nb_machines)
				destination=0;
		    }
		    proc_rank_binome = listeCompacte_alea (& pl_temp.liste_procs, destination, rand_buf);

                    /* on echange ces 2 procs */
                    //DEBUG_PRINT( "swap de %d sur %d avec %d sur %d\n",
		//		  proc_rank, origine,
		//		  proc_rank_binome, destination );

                    //rc = swap_procs( &pl_temp, destination, proc_rank_binome, origine, proc_rank, false );
		    if (check_all_constraints_swap(&pl_temp, proc_rank, origine, proc_rank_binome, destination) &&
			 ( (gain=calcule_gain_swap(initial,&pl_temp,proc_rank,origine,proc_rank_binome, destination))>gainMin)) {
				echange_procs(initial, &pl_temp, proc_rank, origine,proc_rank_binome, destination);
				gainTotal +=gain;
			DEBUG_PRINT("Echange %d  sur %d avec %d sur %d, gain=%ld\n", proc_rank, origine, proc_rank_binome, destination, gain);
		    }

                }
            }  /* end for m */

            /* Is there any overloaded machine? */
/*
            if( !une_machine_est_surchargee(&pl_temp) ){

                new_cout = cout( initial, &pl_temp );

                if( new_cout < cout_init ) {
                    DEBUG_PRINT( "Nouveau cout: %u ; ancien : %u\n", new_cout, cout_init );
                    DEBUG_PRINT( "Amélioration possible: on vérifie ce placement:" );
		    if (check_all_constraints(initial, &pl_temp)) {
			DEBUG_PRINT("\n****** Amélioration ! on garde ce placement.\n");
                    	copier_assignement( &pl_temp, resultat );
		    } else {
			   DEBUG_PRINT("raté...\n");
		    }
                } else {
                    copier_assignement( resultat, &pl_temp);
                }
            }
*/
        } /* end for i */
    } /* end for j */

	copier_assignement(& pl_temp, resultat);
    resultat->cout -= gainTotal;
	DEBUG_PRINT ("Gain total explo_voisinageRand:%ld\n",gainTotal);
    return FMR_OK;
}


/* Calcul du cout total des deplacements des processus */

unsigned long cout_deplacements( unsigned short* to_move ) {
    unsigned long cout_total_depl;
    unsigned short k;
    cout_total_depl = 0;
    for( k = 0 ; k < VOISINAGEMAX ; k++ ) {
        if( to_move[k] != NOPROCESS ) {
            cout_total_depl += instance.procs[to_move[k]].PMC;
        }
    }
    return cout_total_depl;
}

/* Existe-t-il dans le systeme une machine surchargee ? */

bool une_machine_est_surchargee(placement_t* placmnt) {

    unsigned short m, r;

    for( m = 0 ; m < instance.nb_machines ; m++ ) {
        for( r = 0 ; r < instance.nb_ressources ; r++ ) {
            if( placmnt->U[m][r] > instance.machines[m].C[r] ){
                return true;
            }
        }
    }
    return false;
}

/* Retourne le rang de la premiere machine surchargee,
 * -1 si aucune machine n'est surchargee */
/* Il serait peut intéressant de remplacer C[r] par SC[r]*/
unsigned int machine_surchargee(placement_t* placmnt) {
    unsigned short  m, r;

    for( m = 0 ; m < instance.nb_machines ; m++ ) {
        for( r = 0 ; r < instance.nb_ressources ; r++ ) {
            if( placmnt->U[m][r] > instance.machines[m].C[r] ){
                return m;
            }
        }
    }
    return NOMACHINE;
}
