/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include "fmr.h"
#include "fmr_tools.h"
#include "fmr_cost.h"
#include "fmr_constraints.h"
#include "fmr_psmvt.h"
#include "fmr_heap.h"
#include "fmr_threads.h"
#include "fmr_exploSwap.h"

long
explo_voisinageSwap (const placement_t const *initial, placement_t * resultat,
		     long gainMin)
{
    unsigned short p1, m1, p2, m2;
    long gain, cumulgain = 0;

    for (p1 = 0; p1 < instance.nb_procs; p1++) {
	m1 = resultat->assignment[p1];
	for (p2 = (unsigned short)(p1 + 1); p2 < instance.nb_procs; p2++) {	/* donc p2!=p1 */
	    m2 = resultat->assignment[p2];
	    if (m1 != m2) {	/* on impose que p1 et p2 sont sur des machines différentes */

		if (check_all_constraints_swap (resultat, p1, m1, p2, m2)) {
		    if ((gain =
			 calcule_gain_swap (initial, resultat, p1, m1, p2,
					    m2)) > gainMin) {
			echange_procs (initial, resultat, p1, m1, p2, m2);
			DEBUG_PRINT
			    ("Swap de %3d(%3d) sur %3d et de %3d(%3d) sur %3d, gain=%ld\n",
			     p1, instance.procs[p1].service, m1, p2,
			     instance.procs[p2].service, m2, gain);
			atomic_set_meilleur_placement(resultat,'e');
			cumulgain += gain;
			m1 = m2;
		    }
		}
	    }
	}			// for p2
    }				// for p1
    resultat->cout -= cumulgain;
    DEBUG_PRINT ("Gain total expo_voisinageSwap:%ld\n", cumulgain);
    return (cumulgain);
}

long 
explo_voisinageSwap_restreint(const placement_t const * initial, placement_t *resultat, unsigned short present_ou_indice[], long gainMin) 
{
    unsigned short p1, m1, p2, m2, indice;
    long gain, cumulgain = 0;


    for (p1 = 0; p1 < instance.nb_procs; p1++) {
	if ((indice=present_ou_indice[p1])!=NOPROCESS) { // ps oté par le glouton partiel, on saute à l'indice suivant
		if (indice ==0) { // il n'y a plus de ps non oté
			break;
		} else 
			p1=indice;
	} // sinon il est bien présent.
	m1 = resultat->assignment[p1];
	for (p2 = (unsigned short)(p1 + 1); p2 < instance.nb_procs; p2++) {	/* donc p2!=p1 */
	    if ((indice=present_ou_indice[p2])!=NOPROCESS) {
		if (indice ==0)
			break;
		else
			p2=indice;
	    }
	    m2 = resultat->assignment[p2];
	    if (m1 != m2) {	/* on impose que p1 et p2 sont sur des machines différentes */

		if (check_all_constraints_swap (resultat, p1, m1, p2, m2)) {
		    if ((gain =
			 calcule_gain_swap (initial, resultat, p1, m1, p2,
					    m2)) > gainMin) {
			echange_procs (initial, resultat, p1, m1, p2, m2);
			DEBUG_PRINT
			    ("Swap de %3d(%3d) sur %3d et de %3d(%3d) sur %3d, gain=%ld\n",
			     p1, instance.procs[p1].service, m1, p2,
			     instance.procs[p2].service, m2, gain);
			cumulgain += gain;
			m1 = m2;
		    }
		}
	    }
	}			// for p2
    }				// for p1
    resultat->cout -= cumulgain;
    DEBUG_PRINT ("Gain total expo_voisinageSwap:%ld\n", cumulgain);
    return (cumulgain);
}

long
explo_voisinageVeryBestSwap (const placement_t const *initial,
			     placement_t * resultat, long gainMin)
{
    unsigned short p1, m1, p2, m2;
    unsigned short p1Best = NOPROCESS, m1Best = NOMACHINE, p2Best =
	NOPROCESS, m2Best = NOMACHINE;
    long gain, cumulgain = 0, bestGain = gainMin;

    for (p1 = 0; p1 < instance.nb_procs; p1++) {
	m1 = resultat->assignment[p1];
	for (p2 = (unsigned short)(p1 + 1); p2 < instance.nb_procs; p2++) {	/* donc p2!=p1 */
	    m2 = resultat->assignment[p2];
	    if (m1 != m2) {	/* on impose que p1 et p2 sont sur des machines différentes */
		if (check_all_constraints_swap (resultat, p1, m1, p2, m2)) {
		    if ((gain =
			 calcule_gain_swap (initial, resultat, p1, m1, p2,
					    m2)) > bestGain) {
			bestGain = gain;
			p1Best = p1;
			p2Best = p2;
			m1Best = m1;
			m2Best = m2;
			//DEBUG_PRINT("Amelioration bestGain=%ld, p1=%d,p2=%d\n",bestGain,p1,p2);
		    }
		}
	    }
	}			// for p2
    }				// for p1

    if (p1Best != NOPROCESS) {
	echange_procs (initial, resultat, p1Best, m1Best, p2Best, m2Best);
	DEBUG_PRINT
	    ("Swap de %3d(%3d) sur %3d et de %3d(%3d) sur %3d, gain=%ld\n",
	     p1Best, instance.procs[p1Best].service, m1Best, p2Best,
	     instance.procs[p2Best].service, m2Best, bestGain);

    } else
	bestGain = 0;

    cumulgain = bestGain;
    resultat->cout -= bestGain;
    DEBUG_PRINT ("Gain total expo_voisinageBestSwap:%ld\n", cumulgain);
    return (cumulgain);
}

long
explo_voisinageVeryBestSwap_restreint (const placement_t const *initial,
			     placement_t * resultat, unsigned short present_ou_indice[], long gainMin)
{
    unsigned short p1, m1, p2, m2, indice;
    unsigned short p1Best = NOPROCESS, m1Best = NOMACHINE, p2Best =
	NOPROCESS, m2Best = NOMACHINE;
    long gain, cumulgain = 0, bestGain = gainMin;

    for (p1 = 0; p1 < instance.nb_procs; p1++) {
	if ((indice=present_ou_indice[p1])!=NOPROCESS) { // ps oté par le glouton partiel, on saute à l'indice suivant
		if (indice ==0) { // il n'y a plus de ps non oté
			break;
		} else 
			p1=indice;
	} // sinon il est bien présent.
	m1 = resultat->assignment[p1];
// 	printf("VeryBestSwap p1=%d m1=%d\n", p1, m1);
	for (p2 = (unsigned short)(p1 + 1); p2 < instance.nb_procs; p2++) {	/* donc p2!=p1 */
	    if ((indice=present_ou_indice[p2])!=NOPROCESS) {
		if (indice ==0)
			break;
		else
			p2=indice;
	    }
	    m2 = resultat->assignment[p2];
	    if (m1 != m2) {	/* on impose que p1 et p2 sont sur des machines différentes */
		if (check_all_constraints_swap (resultat, p1, m1, p2, m2)) {
		    if ((gain =
			 calcule_gain_swap (initial, resultat, p1, m1, p2,
					    m2)) > bestGain) {
			bestGain = gain;
			p1Best = p1;
			p2Best = p2;
			m1Best = m1;
			m2Best = m2;
			//DEBUG_PRINT("Amelioration bestGain=%ld, p1=%d,p2=%d\n",bestGain,p1,p2);
		    }
		}
	    }
	}			// for p2
    }				// for p1

    if (p1Best != NOPROCESS) {
	echange_procs (initial, resultat, p1Best, m1Best, p2Best, m2Best);
	DEBUG_PRINT
	    ("Swap de %3d(%3d) sur %3d et de %3d(%3d) sur %3d, gain=%ld\n",
	     p1Best, instance.procs[p1Best].service, m1Best, p2Best,
	     instance.procs[p2Best].service, m2Best, bestGain);

    } else
	bestGain = 0;

    cumulgain = bestGain;
    resultat->cout -= bestGain;
    DEBUG_PRINT ("Gain total expo_voisinageBestSwap:%ld\n", cumulgain);
    return (cumulgain);
}


/**
 * Algo qui explore un voisinage par swap (distance 1), on ne garde que les déplacemnts
 * valides dans une Heap et à la fin on ne fait que le déplacement du meilleur 
 */
long
explo_voisinageBestSwap (const placement_t const *initial,placement_t * resultat, long gainMin)
{
    unsigned short p1, p2, m1, m2, i;
    long gain, gainTotal = 0;
    unsigned perdus=0;

    /* Arbre binaire quasi complet pour stocker les meilleurs mouvements,
       seul le loadcost est considéré (semble le critère prépondérant).
       tabGain sera donc ordonné */

    deplt_t **tabGain ;

    //DEBUG_STATS_PRINT("memoire restante avant explo_voisinageBestSwap:%lu, on va consommer: ", memoireRestante());

    if ((unsigned long)MAXNBDEPL < (unsigned long)(instance.nb_procs * (instance.nb_procs-1)/2)) {
            tabGain = (deplt_t **) malloc (MAXNBDEPL * sizeof (deplt_t *));
    } else {
            tabGain = (deplt_t **) malloc ((size_t)instance.nb_procs * ((size_t)instance.nb_procs-1)/2 *sizeof (deplt_t *));
    }

    if (tabGain==NULL) {
	    DEBUG_PRINT("malloc returns NULL\n");
	    return((long)0);
    }
    unsigned nbEltTabGain = 0;
    bool *dejaVu = (bool *) calloc (instance.nb_procs, sizeof (bool));

    for (p1 = 0; p1 < instance.nb_procs; p1++) {
	m1 = resultat->assignment[p1];
	for (p2 = (unsigned short)(p1 + 1); p2 < instance.nb_procs; p2++) {	/* donc p2!=p1 */
	    m2 = resultat->assignment[p2];
	    if (m1 != m2) {	/* on impose que p1 et p2 sont sur des machines différentes */
		if (check_all_constraints_swap (resultat, p1, m1, p2, m2)) {
			if ((gain=calcule_gain_swap (initial, resultat, p1, m1,p2, m2))>gainMin)
		          ajoutHeap (p1, p2, gain, tabGain, &nbEltTabGain);
		}
	    }
	}			// for p2
    }				// for p1

    DEBUG_STATS_PRINT("Nbre de deplacement swap trouvés:%u/%u\n", nbEltTabGain, MAXNBDEPL);
    /* ON récupère les valeurs du Heap en ordre décroissant, en notant les ps dejavus */
    while (nbEltTabGain > 0 && tabGain[0]->gainLoad > gainMin) {
	m1 = resultat->assignment[tabGain[0]->u1];
	m2 = resultat->assignment[tabGain[0]->u2];
	if (!dejaVu[tabGain[0]->u1] && !dejaVu[tabGain[0]->u2]
	    && check_all_constraints_swap (resultat, tabGain[0]->u1, m1,
					   tabGain[0]->u2, m2)
	    && (gain = calcule_gain_swap (initial, resultat, tabGain[0]->u1, m1,
				   tabGain[0]->u2, m2)) > gainMin) {
	    DEBUG_PRINT ("swap de %u sur %u et %u sur %u, gain=%ld\n",
			 tabGain[0]->u1, m1, tabGain[0]->u2, m2, gain);

	    echange_procs (initial, resultat, tabGain[0]->u1, m1,
			   tabGain[0]->u2, m2);
	    gainTotal += gain;
	    dejaVu[tabGain[0]->u1] = true;
	    dejaVu[tabGain[0]->u2] = true;
	    atomic_set_meilleur_placement(resultat,'w');
	} else {
		perdus++;
	}

	oteHeap (tabGain, &nbEltTabGain);
    }

    resultat->cout -= gainTotal;
    DEBUG_STATS_PRINT("perdus=%u\n",perdus);
    DEBUG_PRINT ("Gain total explo_voisinageBestShift:%ld\n", gainTotal);
    free_heap_elts (tabGain, nbEltTabGain);
    free (tabGain);
    free (dejaVu);
    return gainTotal;
}

long
explo_voisinageBestSwap_restreint (const placement_t const *initial,placement_t * resultat, unsigned short present_ou_indice[], long gainMin)
{
    unsigned short p1, p2, m1, m2, i, indice;
    long gain, gainTotal = 0;

    /* Arbre binaire quasi complet pour stocker les meilleurs mouvements,
       seul le loadcost est considéré (semble le critère prépondérant).
       tabGain sera donc ordonné */

    deplt_t **tabGain ;

	printf("memoire restante avant explo_voisinageBestSwap:%lu, on va consommer: ", memoireRestante());

    if ((unsigned long)MAXNBDEPL < (unsigned long)(instance.nb_procs * (instance.nb_procs-1)/2)) {
            tabGain = (deplt_t **) malloc (MAXNBDEPL * sizeof (deplt_t *));
	printf("%lu\n",MAXNBDEPL * (sizeof (deplt_t *)+sizeof(deplt_t)));
    } else {
            tabGain = (deplt_t **) malloc ((size_t)instance.nb_procs * ((size_t)instance.nb_procs-1)/2 *sizeof (deplt_t *));
	printf("%lu\n",instance.nb_procs * ((size_t)instance.nb_procs-1)/2 *(sizeof (deplt_t *)+sizeof(deplt_t)));
    }

    if (tabGain==NULL) {
	    DEBUG_PRINT("malloc returns NULL\n");
	    return((long)0);
    }
    unsigned nbEltTabGain = 0;
    bool *dejaVu = (bool *) calloc (instance.nb_procs, sizeof (bool));

    for (p1 = 0; p1 < instance.nb_procs; p1++) {
	if ((indice=present_ou_indice[p1])!=NOPROCESS) { // ps oté par le glouton partiel, on saute à l'indice suivant
		if (indice ==0) { // il n'y a plus de ps non oté
			break;
		} else 
			p1=indice;
	} // sinon il est bien présent.
	m1 = resultat->assignment[p1];
	for (p2 = (unsigned short)(p1 + 1); p2 < instance.nb_procs; p2++) {	/* donc p2!=p1 */
	    if ((indice=present_ou_indice[p2])!=NOPROCESS) {// ps oté par le glouton partiel, on saute à l'indice suivant
		if (indice ==0) // il n'y a plus de ps non oté
			break;
		else
			p2=indice;
	    }
	    m2 = resultat->assignment[p2];
	    if (m1 != m2) {	/* on impose que p1 et p2 sont sur des machines différentes */
		if (check_all_constraints_swap (resultat, p1, m1, p2, m2)) {
			if ((gain=calcule_gain_swap (initial, resultat, p1, m1,p2, m2))>gainMin)
		          ajoutHeap (p1, p2, gain, tabGain, &nbEltTabGain);
		}
	    }
	}			// for p2
    }				// for p1

    /* ON récupère les valeurs du Heap en ordre décroissant, en notant les ps dejavus */
    while (nbEltTabGain > 0 && tabGain[0]->gainLoad > gainMin) {
	m1 = resultat->assignment[tabGain[0]->u1];
	m2 = resultat->assignment[tabGain[0]->u2];
	if (!dejaVu[tabGain[0]->u1] && !dejaVu[tabGain[0]->u2]
	    && check_all_constraints_swap (resultat, tabGain[0]->u1, m1,
					   tabGain[0]->u2, m2)
	    && (gain = calcule_gain_swap (initial, resultat, tabGain[0]->u1, m1,
				   tabGain[0]->u2, m2)) > gainMin) {
	    DEBUG_PRINT ("swap de %u sur %u et %u sur %u, gain=%ld\n",
			 tabGain[0]->u1, m1, tabGain[0]->u2, m2, gain);

	    echange_procs (initial, resultat, tabGain[0]->u1, m1,
			   tabGain[0]->u2, m2);
	    gainTotal += gain;
	    dejaVu[tabGain[0]->u1] = true;
	    dejaVu[tabGain[0]->u2] = true;
	}
	oteHeap (tabGain, &nbEltTabGain);
    }

    resultat->cout -= gainTotal;
    DEBUG_PRINT ("Gain total explo_voisinageBestShift:%ld\n", gainTotal);
    free_heap_elts (tabGain, nbEltTabGain);
    free (tabGain);
    free (dejaVu);
    return gainTotal;
}
