/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include "fmr_recuit.h"

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <stdbool.h>

#include "fmr.h"
#include "fmr_mem.h"
#include "fmr_cost.h"
#include "fmr_tools.h"
#include "fmr_tirage.h"
#include "fmr_threads.h"

#define min(X,Y) ((X<Y) ? (X) : (Y))
#define max(X,Y) ((X>Y) ? (X) : (Y))

static long
recuit (const placement_t * initial, placement_t * resultat,
	double taux_transitions_initial, double TFinale, double alpha,
	unsigned int NbitPalier, unsigned int Nbloque, struct drand48_data * rand_buf);




/**
 * API
 */
long 
explo_recuit_best_loop(const placement_t* initial, struct drand48_data* rand_buf)
{
    placement_t resultatc;
    resultatc = alloc_assignement (instance.nb_procs);
    while (true)
    {
	atomic_if_meilleur_placement_copyto( &resultatc );
	explo_recuit(initial, &resultatc, rand_buf);
    }
}

long
explo_recuit (const placement_t const *initial, placement_t * courante, struct drand48_data * rand_buf)
{
//     return recuit( initial, courante, (double)courante->cout, 0.001, 0.975, 100000, 10000);
//      return recuit( initial, courante, (double)initial->cout, 0.001, 0.975, 100000, 10000);
/*     return recuit( initial, courante, 100000., 0.001, 0.975, 100000, 10000);*/
//      return recuit( initial, courante, temperature_initiale(initial,courante,0.8,10000), 0.001, 0.975, 100000, 10000);
    return recuit (initial, courante, 0.3, 0.001, 0.975, 10000, 10000, rand_buf);
}


/**
 * Recuit simulé

Données:
    Ri        : une affectation des processus
    Tx        : taux initial souhaité de transitions acceptées
    Tf        : température finale
    alpha     : coefficient de refroidissement
    NbItBase  : nombre d'itérations prévu pour chaque pallier de température
    NbItGel   : nombre de tirages consécutifs infructueux après lesquels la situation est dite "gelée"


Algo:
   Première phase : recherche d'une température initiale Ti

   Ti <- Tf
   R <- Ri             (R = affectation des processus dans la solution "courante")
   Répéter :
   |    Ti <- Ti * 2
   |    NbDepPallier <- 0   (NbDepPallier = nombre de réaffectations de R pour la température courante )
   |    NbIt <- NbItBase
   |    Répéter NbIt fois :
   |    |   Essayer au plus NBitGel fois, jusqu'à tomber sur un mouvement réalisable :
   |    |       Rtemp <- Tirage aléatoire d'un mouvement (de type swap ou shift)
   |    |   Si aucun mouvement tiré n'est réalisable :
   |    |       Sortir du recuit (système gelé)
   |    |   R <- R ou Rtemp ( choix aléatoire entre les deux, selon des probabilités proportionnelles )
   |    |                   ( à ( exp(-cout(R)/Ti) , exp(-cout(Rtemp)/Ti) )                           )
   |    |   Si Rtemp a été choisi :
   |    |        NbDepPallier <- NbDepPallier + 1
   |    |   Si R améliore la meilleure solution connue :
   |    |_       NbIt <- NbIt + nbItBase
   |
   |_Jusqu'à (NbDepPallier/NbIt > Tx)
   

   Seconde phase : descente de température

   Pour T allant de Ti à Tf (par multiplication par alpha) :
   |    NbIt <- NbItBase
   |    Répéter NbIt fois :
   |    |   Essayer au plus NBitGel fois, jusqu'à tomber sur un mouvement réalisable :
   |    |       Rtemp <- Tirage aléatoire d'un mouvement (de type swap ou shift)
   |    |   Si aucun mouvement tiré n'est réalisable :
   |    |       Sortir du recuit (système gelé)
   |    |   R <- R ou Rtemp ( choix aléatoire entre les deux, selon des probabilités proportionnelles )
   |    |                   ( à ( exp(-cout(R)/T) , exp(-cout(Rtemp)/T) )                             )
   |    |   Si R améliore la meilleure solution connue :
   |    |_       NbIt <- NbIt * 2
   |_

 */


static long
recuit (const placement_t const *initial, placement_t * resultat,
	double taux_transitions_initial, double TFinale, double alpha,
	unsigned NbitPalierBase, unsigned Nbloque, struct drand48_data * rand_buf)
{
    unsigned p, m_orig = 0, m, nbDep = 0, bloque =
	0, ItPalier, nbDepPalier, nbItPalier;
    int64 Delta, gainTotal;
    int64 initCout;
    double T;			// Temperature courante
    placement_t resultatc;
    tirage_t tirage;		// tirage (aléatoire ou pas) d'un mouvement
    int phase_cherche_Tinit = 1;

#ifdef EXAMINE_RECUIT
    int64 pallier_nb_essais;	// nombre d'essais avant de trouver un tirage possible (< Nbloque) + = +contraint
    int64 pallier_nb_transitions;	// nombre de transitions effectuées sur un pallier (< NbitPalier) + = -contraint
    int64 pallier_nb_meilleurs;	// nombre de tirages trouvés qui améliorent le placement courant (< NbitPalier) + = +loin de l'optimum local
    int64 pallier_cout_min;	// cout min des configurations par lesquelles on passe
    int64 pallier_cout_max;	// cout max des configurations par lesquelles on passe
    int64 pallier_cout_max_tirage;	// cout max des tirages d'un pallier
#endif

    m = m_orig;
    initCout = resultat->cout;	// cout initial
    resultatc = alloc_assignement (instance.nb_procs);
    copier_assignement (resultat, &resultatc);

//     PRINT_EXAMINE_RECUIT("Température\tCoût\n");

    T = 2 * TFinale;
    while (T > TFinale)		// Boucle Principale
    {

	ItPalier = nbDepPalier = 0;
	nbItPalier = NbitPalierBase;

#ifdef EXAMINE_RECUIT
	pallier_nb_meilleurs = pallier_nb_transitions = pallier_nb_essais = 0;
	pallier_cout_min = resultatc.cout;
	pallier_cout_max = resultatc.cout;
	pallier_cout_max_tirage = resultatc.cout;
#endif

//         copier_assignement(resultat, &resultatc);

	while (ItPalier < nbItPalier)	// NbitPalier Iterations pour une temperature fixee
	{
	    pthread_testcancel();

	    // Transition elementaire (valide) (shift) choisie aleatoirement
	    // Nbloque permet de sortir au cas où aucune transition valide n'est trouvee

	    tirage = tirage_next (&resultatc, Nbloque, rand_buf);
#ifdef EXAMINE_RECUIT
	    pallier_nb_essais += tirage.nb_essais;
#endif

	    if (tirage.nouveau)	// transition valide trouvee
	    {
		Delta = calcule_gain_tirage (initial, &resultatc, tirage);

#ifdef EXAMINE_RECUIT
		if (Delta > 0)
		    pallier_nb_meilleurs++;
		pallier_cout_max_tirage =
		    max (pallier_cout_max_tirage, resultatc.cout - Delta);
#endif

		// si gain on bouge systematiquement, sinon ça dépend de la température
//                 if (Delta>0 || exp((double)Delta/T) > frand_a_b(0.0, 1.0)) // (1) prop. Frédéric
//                 if (Delta>0 || exp(-1+(double)Delta/T) > frand_a_b(0.0, 1.0)) // (2) svn r.244 : prop. Fédéric modifiée pour que Delta=0 => prob=0.5 (mais formule fausse) 
//                 if (Delta>0 || exp((double)Delta/T) > frand_a_b(0.0, 2.0)) // (3) prop. Frédéric modifiée pour que Delta=0 => prob=0.5
		if (frand_a_b (rand_buf, -1.0, exp ((double) Delta / T)) > 0)	// (4) prop. Antoine
//                 if (Delta>0 || frand_a_b(-1.0, exp((double)Delta/T))>0) // (5) prop. Antoine modifiée
//                 if (Delta>0 || (1+(double)Delta/T) > frand_a_b(0.0, 2.0)) // (3bis) linéarisation --> plus lent !
		{
		    nbDep++;
		    nbDepPalier++;
		    effectue_tirage (initial, &resultatc, tirage);
		    resultatc.cout -= Delta;

#ifdef EXAMINE_RECUIT
		    pallier_nb_transitions++;
		    pallier_cout_min = min (pallier_cout_min, resultatc.cout);
		    pallier_cout_max = max (pallier_cout_max, resultatc.cout);
#endif

		    if (resultatc.cout < resultat->cout) {
			
			copier_assignement (&resultatc, resultat);	// et on sauve si c'est la meilleure solution trouvee : A OPTIMISER si possible !
			atomic_set_meilleur_placement(resultat,'a');
			
			DEBUG_PRINT
			    ("gain                                          : %lld\n",
			     (long long int) Delta);
			DBG_AFFICHE_TIRAGE (tirage, resultat->cout);
			DEBUG_PRINT
			    ("verif                                         : %11lld\n",
			     cout (initial, resultat));
			DEBUG_PRINT
			    ("                                                T : %11f\n",
			     T);
			DEBUG_PRINT
			    ("                                        itPallier : %11u\n",
			     ItPalier);

			/* si amélioration globale, on reste un peu plus longtemps sur ce pallier */
			if (!phase_cherche_Tinit
			    && ItPalier > nbItPalier - NbitPalierBase)
			    nbItPalier += NbitPalierBase;


//                      DEBUG_PRINT("nbItPalier = %u\n", nbItPalier);
		    }
		    /*                   else
		       { **
		       DEBUG_PRINT("gain (globalement négatif)                    : %11ld\n", Delta);
		       DBG_AFFICHE_TIRAGE(tirage, resC);
		       DEBUG_PRINT("verif                                         : %11ld\n", cout(initial, &resultatc));
		       DEBUG_PRINT("                                                T : %11f\n", T);
		       DEBUG_PRINT("                                        itPallier : %11u\n", ItPalier);
		       } */
		}
	    } else		// le systeme est "gele" car aucune transition valide n'a ete trouvee : on arrete les frais...
	    {
		nbItPalier = ItPalier - 1;
		if (!phase_cherche_Tinit)
		    T = TFinale;
	    }

	    ItPalier++;
	}			// Fin palier de temperature

	PRINT_EXAMINE_RECUIT("%11f\t%11lld\t%11lld\t%11lld\t%11lld\t%11lld\t%11lld\t%11lld\n",
	     T, resultatc.cout, pallier_cout_min, pallier_cout_max,
	     pallier_cout_max_tirage, pallier_nb_essais,
	     pallier_nb_transitions, pallier_nb_meilleurs);

	/* augmentation de la température jusqu'à atteindre un taux de transition suffisant */
	if (phase_cherche_Tinit
	    && nbDepPalier < taux_transitions_initial * nbItPalier)
	    T *= 2;
	/* ensuite, diminution de la température */
	else {
	    phase_cherche_Tinit = false;
	    T *= alpha;		// on baisse la temperature
	}
    }				// Fin Boucle principale

    gainTotal = (initCout - resultat->cout);

    DEBUG_PRINT ("Nbre deplacements :%u\n", nbDep);
    DEBUG_PRINT
	("Coût initial:%lld\tGain total Recuit:%lld\t Coût final:%lld\n",
	 (long long int) initCout, (long long int) gainTotal, resultat->cout);
	
    free_placement(&resultatc);
    return (long) gainTotal;
}
