/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#define  DELAI_CALCUL_TEMPERATURE_INITIALE 30.
#define  MAX_TIRAGES_TEMPERATURE_INITIALE  10000
#define  DELAI_UNE_PASSE                   60.
#define  TEMPERATURE_FINALE                 1.
#define  DELAI_TIRAGES_INVALIDES            5.
#define  DELAI_GEL                          10.
#define  LINEAIRE 1
#define  CARREE   2
#define  FONCTION_TEMPERATURE               CARREE

#include "fmr_recuit_b.h"

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <stdbool.h>

#include "fmr.h"
#include "fmr_mem.h"
#include "fmr_cost.h"
#include "fmr_tools.h"
#include "fmr_tirage.h"
#include "fmr_threads.h"

#define min(X,Y) ((X<Y) ? (X) : (Y))
#define max(X,Y) ((X>Y) ? (X) : (Y))


/*
 * calcul de la température en fonction du temps
 */
temperature_t
  temperature (temperature_t);


/*
 * initialisation de la structure temperature_t
 */
temperature_t init_recuit (double T0, double T1);

/*
 * calcul de la température initiale
 */
double
temperature_initiale (const placement_t * initial,
                      placement_t * courante,
                      struct drand48_data *rand_buf);



/* *******************************************************************
 *                              API
 * *******************************************************************/

/**
 * boucle sans fin relançant le recuit sur le meilleur résultat connu
 */
long
explo_recuit_best_loop_b (const placement_t * initial,
                          struct drand48_data *rand_buf)
{
    placement_t resultatc;
    resultatc = alloc_assignement (instance.nb_procs);
    atomic_if_meilleur_placement_copyto (&resultatc);
    double T0 = temperature_initiale (initial, &resultatc, rand_buf);

    while (true) {
        atomic_if_meilleur_placement_copyto (&resultatc);
        recuit_b (initial, &resultatc, init_recuit (T0, TEMPERATURE_FINALE),
                rand_buf);
    }
}

/**
 * lancement du recuit (une seule passe) à partir d'une solution courante
 */
long
explo_recuit_b (const placement_t const *initial, placement_t * courante,
                struct drand48_data *rand_buf)
{
    double T0 = temperature_initiale (initial, courante, rand_buf);
    return recuit_b (initial, courante, init_recuit (T0, TEMPERATURE_FINALE),
                   rand_buf);
}


/**
 * Recuit simulé

Données:
    Ri        : une affectation des processus
    


Algo:
   Première phase : recherche d'une température initiale Ti
   
   Tirer aléatoirement 10000 mouvements possibles depuis la solution initiale (ou temps max DELAI_CALCUL_TEMPERATURE_INITIALE)
   Stocker les valeurs absolues des gains pour chaque mouvement
   Choisir la température initiale Ti comme étant la valeur médiane de ces valeurs absolues (ce qui donne environ 50% d'acceptation)


   Seconde phase : descente de température

   Pendant DELAI_UNE_PASSE secondes :
   |    Calculer la température T=f(t,Ti,Tf) (t étant le temps)
   |    Tirer aléatoirement un mouvement Rtemp (de type swap ou shift) réalisable
   |    Si aucun mouvement réalisable n'est trouvé en moins de DELAI_TIRAGES_INVALIDES secondes
   |    |   Sortir du crecuit (système bloqué)
   |    R <- R ou Rtemp ( choix aléatoire entre les deux, selon des probabilités proportionnelles )
   |                    ( à ( exp(-cout(R)/T) , exp(-cout(Rtemp)/T) )                             )
   |    Si aucun Rtemp n'a été choisi depuis DELAI_GEL secondes
   |    |   Sortir du recuit (système gelé)
   |_

 */



/*
 * calcul de la température initiale
 */
// to be used by qsort
static int
compare_int64(const void *pa, const void *pb)
{
    const int64 a = *(const int64 *)pa;
    const int64 b = *(const int64 *)pb;
    return (a<b ? -1 : a==b ? 0 : 1);
}

double
temperature_initiale (const placement_t * initial,
                      placement_t * courante,
                      struct drand48_data *rand_buf
                     )
{
    int64 deltas[MAX_TIRAGES_TEMPERATURE_INITIALE];
    size_t   nb_tirages = 0;
    double tmax = tempsEcoule2() + DELAI_CALCUL_TEMPERATURE_INITIALE ;
    tirage_t tirage;            // tirage (aléatoire ou pas) d'un mouvement
    do {
        pthread_testcancel ();
        tirage = tirage_next_b (courante, tmax, rand_buf);
        if (tirage.nouveau)
        {
            // transition valide trouvee
            deltas[nb_tirages++] = labs(calcule_gain_tirage(initial, courante, tirage));
        }
    }
    while (tmax>tempsEcoule2() && nb_tirages<MAX_TIRAGES_TEMPERATURE_INITIALE && tirage.nouveau);
    
    qsort(deltas, nb_tirages, sizeof(int64), compare_int64);
    return (double) deltas[ nb_tirages / 2];
}



long
recuit_b (const placement_t const *initial, placement_t * resultat,
        temperature_t s, struct drand48_data *rand_buf)
{
    unsigned nbDep = 0;
    int64 Delta, gainTotal = 0;
    int64 initCout, bestCout;
    tirage_t tirage;            // tirage (aléatoire ou pas) d'un mouvement
    double limiteGel = tempsEcoule2() + DELAI_GEL ;
    initCout = bestCout = resultat->cout;   // cout initial

//     PRINT_EXAMINE_RECUIT("Température\tCoût\n");

    while (s=temperature(s), s.t < s.t1 && s.t<limiteGel)         // Boucle Principale : tant qu'on a le temps
    {
        //         copier_assignement(resultat, &resultatc);

        pthread_testcancel ();
        
        // Transition elementaire (valide) (shift) choisie aleatoirement
        // Nbloque permet de sortir au cas où aucune transition valide n'est trouvee
        
        tirage = tirage_next_b (resultat, s.t + DELAI_TIRAGES_INVALIDES, rand_buf);
        
        if (tirage.nouveau) // transition valide trouvee
            {
                Delta = calcule_gain_tirage (initial, resultat, tirage);
                
                // si gain on bouge systematiquement, sinon ça dépend de la température
                if (Delta>0 || exp((double)Delta/s.T) > frand_a_b(rand_buf, 0.0, 1.0)) // (1) prop. Frédéric
                {
                    nbDep++;
                    effectue_tirage (initial, resultat, tirage);
                    resultat->cout -= Delta;
                    gainTotal += Delta;
                    limiteGel = tempsEcoule2() + DELAI_GEL ;

                    // on sauve si c'est la meilleure solution trouvee
                    if (resultat->cout < bestCout) {
                        bestCout = resultat->cout;
                        atomic_set_meilleur_placement (resultat, 'a');
                    }
                }
            } else              // le systeme est "gele" car aucune transition valide n'a ete trouvee : on arrete les frais...
            {
                break;
            }
            
            PRINT_EXAMINE_RECUIT("%13.0f%13lld%5d\n", s.T, resultat->cout, (int)s.t);
    }  // Fin Boucle principale
    

    DEBUG_PRINT ("Nbre deplacements :%u\n", nbDep);
    DEBUG_PRINT
    ("Coût initial:%lld\tGain total Recuit:%lld\t Coût final:%lld\n",
     (long long int) initCout, (long long int) gainTotal, bestCout);
    
    return (long) gainTotal;
}








#if FONCTION_TEMPERATURE==LINEAIRE
temperature_t
temperature (temperature_t s)
{
    s.t = tempsEcoule2();
    double lambda = (s.t1 - s.t) / s.delta_t ;
    s.T = lambda*s.T0 + (1-lambda)*s.T1;
    return s;
}
#endif

#if FONCTION_TEMPERATURE==CARREE
temperature_t
temperature (temperature_t s)
{
    s.t = tempsEcoule2();
    double lambda = (s.t1 - s.t) / s.delta_t ;
    lambda *= lambda ;
    s.T = lambda*s.T0 + (1-lambda)*s.T1;
    return s;
}
#endif

/**
 * initialisation de la structure temperature_t
 */
temperature_t
init_recuit (double T0, double T1)
{
    temperature_t s;
    s.T0 = T0;
    s.T1 = T1;
    s.t0 = tempsEcoule2 ();
    s.t1 = s.t0 + DELAI_UNE_PASSE;
    s.delta_t = s.t1 - s.t0 ;
    s.delta_T = s.T1 - s.T0 ;
    s.t = s.t0;
    s.T = s.T0;
    return s;
}
