/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include "fmr_stats.h"

#include "fmr.h"

#include <math.h>


/**
 * Calcul de la moyenne et de l'ecart-type de la charge des machines
 * pour chaque ressource. Rempli un tableau de stats_t correspondant
 * a chaque ressource.
 * Algo utilise : un truc de Knuth trouve dans un de ses bouquins
 */


stats_t* calcul_stats_placement( placement_t plt ) {

    unsigned m, r;
    double charge, safe, capa;
    double sum, sum_safe;

    stats_t* statistiques;

    statistiques = (stats_t*) malloc( instance.nb_ressources * sizeof( stats_t ) );

    /* on parcourt la liste des machines ressource par ressource
     c'est ignoble niveau acces memoires mais algorithmiquement 
     ca colle mieux avec l'algo de Knuth qui calcule ca en ligne 
    */

    for( r = 0 ; r < instance.nb_ressources ; r++ ){

        charge = plt.U[0][r];
        safe = instance.machines[0].SC[r];
        capa = instance.machines[0].C[r];

        statistiques[r].moyenne = charge / capa;
        sum = ( charge / capa ) * ( charge / capa );

        statistiques[r].moyenne_safe = ( safe - charge ) / safe;
        sum_safe = ( ( safe - charge ) / safe ) * ( ( safe - charge ) / safe );

        statistiques[r].min = charge / capa;
        statistiques[r].max = charge / capa;
        
        for( m = 0 ; m < instance.nb_machines ; m ++ ) {
            charge = plt.U[m][r];
            safe = instance.machines[m].SC[r];
            capa = instance.machines[m].C[r];
	    DEBUG_PRINT( "capa = %lf, safe = %lf, charge = %lf\n", capa, safe, charge );

            if( ( charge / capa ) > statistiques[r].max ) {
                statistiques[r].max = charge / capa;
            } else {
                if( ( charge / capa ) < statistiques[r].min ) {
                    statistiques[r].min = charge / capa;
                } 
            }

            if( m != 0 ) {
                statistiques[r].moyenne /= m;
                statistiques[r].stddev = sum/m;
                statistiques[r].stddev -= ( statistiques[r].moyenne * statistiques[r].moyenne);
                statistiques[r].stddev = sqrt( statistiques[r].stddev );

                statistiques[r].moyenne_safe /= m;
                statistiques[r].stddev_safe = sum_safe/m;
                statistiques[r].stddev_safe -= ( statistiques[r].moyenne_safe * statistiques[r].moyenne_safe );
                statistiques[r].stddev_safe = sqrt( statistiques[r].stddev_safe );
            } 
            
            statistiques[r].moyenne += charge / capa;
            sum += ( charge / capa ) * ( charge / capa );

            statistiques[r].moyenne_safe += ( safe - charge ) / safe;
            sum_safe += ((  safe - charge ) / safe ) * ( ( safe - charge ) / safe );

        }

	statistiques[r].moyenne /= instance.nb_machines;
        statistiques[r].moyenne_safe /= instance.nb_machines;

	statistiques[r].stddev = sum / instance.nb_machines;
	statistiques[r].stddev -= ( statistiques[r].moyenne * statistiques[r].moyenne );
        statistiques[r].stddev = sqrt( statistiques[r].stddev );

	statistiques[r].stddev_safe = sum_safe / instance.nb_machines;
        statistiques[r].stddev_safe -= ( statistiques[r].moyenne_safe * statistiques[r].moyenne_safe );
        statistiques[r].stddev_safe = sqrt( statistiques[r].stddev_safe );
    
    }
    
    return statistiques;

}

/**
 * Afficher les stats qu'on a calculees sur un placement 
 */

int afficher_stats( stats_t* st ) {

    unsigned r;

    if( NULL == st ) {
        return FMR_NULL_PTR;
    }

        DEBUG_STATS_PRINT( " * ** **** Statistiques **** ** *\n" );
    for( r = 0 ; r < instance.nb_ressources ; r++ ){
        DEBUG_STATS_PRINT( " ** Ressource %d **\n", r );
        DEBUG_STATS_PRINT( "Charge moyenne : %lf\tstd dev : %lf\n", st[r].moyenne, st[r].stddev );
        DEBUG_STATS_PRINT( "          Safe : %lf\tstd dev : %lf\n", st[r].moyenne_safe, st[r].stddev_safe );
        DEBUG_STATS_PRINT( "          min  : %lf\t    max : %lf\n", st[r].min, st[r].max );



    }


    return FMR_OK;
}

/**
 * La fonction de Franck qui verifie les depassements de safe capa
*/

void check_depassements_capa(placement_t * placmnt) {
unsigned int r,m,countPerMach=0,count=0;
bool trouve;

        for( m = 0 ; m < instance.nb_machines ; m++ ) {
	    trouve=false;
            for( r = 0 ; r < instance.nb_ressources ; r++ ){
                if (placmnt->U[m][r] < instance.machines[m].SC[r]) {
                        //DEBUG_PRINT("M%dR[%d] ne depasse pas SC[%d/%d]\n",m,r,placmnt->U[m][r], instance.machines[m].SC[r]);
                        count++;
			trouve=true;
                }
            }
	    if (trouve) 
			countPerMach++;
        }
        printf("%u (pour %u mach/%u) non depassements de SC\n",count, countPerMach, instance.nb_machines);
}

