/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include <pthread.h>

#include "fmr_psmvt.h"
#include "fmr_cost.h"
#include "fmr_constraints.h"
#include "fmr_tools.h"
#include "fmr_tirage.h"


/**
*  tirages limités par le nombre d'essais
*/

tirage_t tirage_next(placement_t* placement, unsigned long max_essais, struct drand48_data * rand_buf)
{
    tirage_t tirage;
    unsigned int r = irand_a_b(rand_buf, 0,2);
    unsigned int r0 = r;
    #ifdef EXAMINE_RECUIT
    int64 nb_essais = 0 ;
    #endif
    do {
        switch ( r ) {
        case 0 :
            tirage = tirage_shift_checked(placement, max_essais, rand_buf);
            break;
        case 1 :
            tirage = tirage_swap_checked( placement, max_essais, rand_buf);
            break;
        default:
            DEBUG_PRINT("Erreur : tirage de type %u demandé\n", r);
            exit(-3);
        }
        #ifdef EXAMINE_RECUIT
        nb_essais += tirage.nb_essais ;
        #endif
        r = (r+1) % 2;
    } while ( r!=r0 && tirage.nouveau==false );
    
    #ifdef EXAMINE_RECUIT
    tirage.nb_essais = nb_essais ;
    #endif

    return tirage;
}


tirage_t tirage_shift_checked(placement_t* placement, unsigned long nb_essais, struct drand48_data * rand_buf)
{
    tirage_t tirage ;
    tirage.nouveau = true;
    tirage.type = SHIFT;
    #ifdef EXAMINE_RECUIT
    tirage.nb_essais = (int64) nb_essais ;
    #endif

    do {
// 	pthread_testcancel();
        tirage.t.shift.p = (unsigned short)irand_a_b(rand_buf, 0, instance.nb_procs);
        tirage.t.shift.m = (unsigned short)irand_a_b(rand_buf, 0, instance.nb_machines);
        nb_essais--;
	DEBUG_PRINT ("tirage_shift_checked: tirage.t.shift.p = %d et tirage.t.shift.m=%d\n", tirage.t.shift.p, tirage.t.shift.m);
    } while(  (signed) nb_essais>0 &&
    ( tirage.t.shift.m == placement->assignment[tirage.t.shift.p]
    ||
    !check_all_constraints_shift( placement, tirage.t.shift.p , placement->assignment[tirage.t.shift.p], tirage.t.shift.m)
    )
    );
    if (nb_essais==0) tirage.nouveau = false;

    #ifdef EXAMINE_RECUIT
    tirage.nb_essais -= (int64) nb_essais ;
    #endif
    return tirage;
}

tirage_t tirage_swap_checked(placement_t* placement, unsigned long nb_essais, struct drand48_data * rand_buf)
{
    tirage_t tirage;
    tirage.nouveau = true;
    tirage.type = SWAP;
    #ifdef EXAMINE_RECUIT
    tirage.nb_essais = (int64) nb_essais ;
    #endif
    do {
	/*pthread_testcancel();*/
        tirage.t.swap.p1 = (unsigned short)irand_a_b(rand_buf, 0, instance.nb_procs);
        tirage.t.swap.p2 = (unsigned short)irand_a_b(rand_buf, 0, instance.nb_procs);
        nb_essais--;
	DEBUG_PRINT ("nb_essais=%lu tirage_swap_checked: tirage.t.swap.p1 = %hu sur placement->assignment[tirage.t.swap.p1] = %hu et tirage.t.swap.p2 = %hu sur  placement->assignment[tirage.t.swap.p2] = %hu\n", nb_essais, tirage.t.swap.p1, placement->assignment[tirage.t.swap.p1], tirage.t.swap.p2, placement->assignment[tirage.t.swap.p2] );
    } while ((signed) nb_essais>0 &&
    (placement->assignment[tirage.t.swap.p1]==placement->assignment[tirage.t.swap.p2]
      ||
      !check_all_constraints_swap( placement, tirage.t.swap.p1, placement->assignment[tirage.t.swap.p1], tirage.t.swap.p2, placement->assignment[tirage.t.swap.p2] )
    )
    );
    if (nb_essais==0) tirage.nouveau = false;

    #ifdef EXAMINE_RECUIT
    tirage.nb_essais -= (int64) nb_essais ;
    #endif
    return tirage;
}

/**
 * tirages limités par le temps
 */
tirage_t tirage_shift_checked_b(placement_t* placement, double tmax, struct drand48_data * rand_buf);
tirage_t tirage_swap_checked_b(placement_t* placement, double tmax, struct drand48_data * rand_buf);

tirage_t tirage_next_b(placement_t* placement, double tmax, struct drand48_data * rand_buf)
{
    tirage_t tirage;
    unsigned int r = irand_a_b(rand_buf, 0,2);
    unsigned int r0 = r;
    #ifdef EXAMINE_RECUIT
    int64 nb_essais = 0 ;
    #endif
    do {
        switch ( r ) {
            case 0 :
                tirage = tirage_shift_checked_b(placement, tmax, rand_buf);
                break;
            case 1 :
                tirage = tirage_swap_checked_b( placement, tmax, rand_buf);
                break;
            default:
                DEBUG_PRINT("Erreur : tirage de type %u demandé\n", r);
                exit(-3);
        }
        #ifdef EXAMINE_RECUIT
        nb_essais += tirage.nb_essais ;
        #endif
        r = (r+1) % 2;
    } while ( r!=r0 && tirage.nouveau==false );
    
    #ifdef EXAMINE_RECUIT
    tirage.nb_essais = nb_essais ;
    #endif
    
    return tirage;
}


tirage_t tirage_shift_checked_b(placement_t* placement, double tmax, struct drand48_data * rand_buf)
{
    double t ;
    tirage_t tirage ;
    tirage.nouveau = true;
    tirage.type = SHIFT;
    do {
        //  pthread_testcancel();
        tirage.t.shift.p = (unsigned short)irand_a_b(rand_buf, 0, instance.nb_procs);
        tirage.t.shift.m = (unsigned short)irand_a_b(rand_buf, 0, instance.nb_machines);
        #ifdef EXAMINE_RECUIT
        tirage.nb_essais++;
        #endif
//         DEBUG_PRINT ("tirage_shift_checked: tirage.t.shift.p = %d et tirage.t.shift.m=%d\n", tirage.t.shift.p, tirage.t.shift.m);
    } 
    while( (t=tempsEcoule2(),tmax>t) &&
        ( tirage.t.shift.m == placement->assignment[tirage.t.shift.p]
        ||
        !check_all_constraints_shift( placement, tirage.t.shift.p , placement->assignment[tirage.t.shift.p], tirage.t.shift.m)
        ));
    
    if (t>=tmax) tirage.nouveau = false;
    
    return tirage;
}

tirage_t tirage_swap_checked_b(placement_t* placement, double tmax, struct drand48_data * rand_buf)
{
    double t ;
    tirage_t tirage;
    tirage.nouveau = true;
    tirage.type = SWAP;
    do {
        /*pthread_testcancel();*/
        tirage.t.swap.p1 = (unsigned short)irand_a_b(rand_buf, 0, instance.nb_procs);
        tirage.t.swap.p2 = (unsigned short)irand_a_b(rand_buf, 0, instance.nb_procs);
        #ifdef EXAMINE_RECUIT
        tirage.nb_essais++;
        #endif
//         DEBUG_PRINT ("temps=%.1f tirage_swap_checked: tirage.t.swap.p1 = %hu sur placement->assignment[tirage.t.swap.p1] = %hu et tirage.t.swap.p2 = %hu sur  placement->assignment[tirage.t.swap.p2] = %hu\n", t, tirage.t.swap.p1, placement->assignment[tirage.t.swap.p1], tirage.t.swap.p2, placement->assignment[tirage.t.swap.p2] );
    } 
    while ((t=tempsEcoule2(),tmax>t) &&
        (placement->assignment[tirage.t.swap.p1]==placement->assignment[tirage.t.swap.p2]
        ||
        !check_all_constraints_swap( placement, tirage.t.swap.p1, placement->assignment[tirage.t.swap.p1], tirage.t.swap.p2, placement->assignment[tirage.t.swap.p2] )
        ));
    
    if (t>=tmax) tirage.nouveau = false;
    
    return tirage;
}



/**
 * tirages limités par le temps, machine cible optimale (non aléatoire)
 */
tirage_t tirage_shift_checked_c(const placement_t* initial, placement_t* placement, double tmax, struct drand48_data * rand_buf);
tirage_t tirage_swap_checked_c(const placement_t* initial, placement_t* placement, double tmax, struct drand48_data * rand_buf);

tirage_t tirage_next_c(const placement_t* initial, placement_t* placement, double tmax, struct drand48_data * rand_buf, tirage_type_t type)
{
    tirage_t tirage;
    
    switch ( type ) {
        case SHIFT :
            tirage = tirage_shift_checked_c(initial, placement, tmax, rand_buf);
            break;
        case SWAP :
            tirage = tirage_swap_checked_c(initial, placement, tmax, rand_buf);
            break;
        default:
            DEBUG_PRINT("Erreur : tirage de type %u demandé\n", r0);
            exit(-3);
    }
    
    return tirage;
}


tirage_t tirage_shift_checked_c(const placement_t* initial, placement_t* placement, double tmax, struct drand48_data * rand_buf)
{
    double t ;
    int64_t gain, gainMax=0 ;
    tirage_t tirage  = {gain:0};
    tirage.nouveau = true;
    tirage.type = SHIFT;
    short unsigned best_m;
    do {
        best_m = NOMACHINE;
        //  pthread_testcancel();
        tirage.t.shift.p = (unsigned short)irand_a_b(rand_buf, 0, instance.nb_procs);
        tirage.t.shift.m = NOMACHINE;
        #ifdef EXAMINE_RECUIT
        tirage.nb_essais++;
        #endif
        for (tirage.t.shift.m=0; tirage.t.shift.m<instance.nb_machines; tirage.t.shift.m++) {
            if (tirage.t.shift.m!=placement->assignment[tirage.t.shift.p]
                && check_all_constraints_shift( placement, tirage.t.shift.p , placement->assignment[tirage.t.shift.p], tirage.t.shift.m)
                && ((gain=calcule_gain_tirage (initial, placement, tirage)) > gainMax))
            {
                best_m = tirage.t.shift.m;
                gainMax = gain;
            }
        }
        //         DEBUG_PRINT ("tirage_shift_checked: tirage.t.shift.p = %d et tirage.t.shift.m=%d\n", tirage.t.shift.p, tirage.t.shift.m);
    } 
    while((t=tempsEcoule2(),tmax>t) && (best_m == NOMACHINE));

    tirage.t.shift.m = best_m;
    tirage.gain = gainMax;
    if (t>=tmax) tirage.nouveau = false;
    
    return tirage;
}

tirage_t tirage_swap_checked_c(const placement_t* initial, placement_t* placement, double tmax, struct drand48_data * rand_buf)
{
    double t ;
    int64_t gain, gainMax=0 ;
    tirage_t tirage = {gain:0};
    tirage.nouveau = true;
    tirage.type = SWAP;
    short unsigned int best_p;
    do {
        best_p = NOPROCESS;
        /*pthread_testcancel();*/
        tirage.t.swap.p1 = (unsigned short)irand_a_b(rand_buf, 0, instance.nb_procs);
        tirage.t.swap.p2 = NOPROCESS;
        unsigned int m1 = placement->assignment[tirage.t.swap.p1];
        #ifdef EXAMINE_RECUIT
        tirage.nb_essais++;
        #endif
        for (tirage.t.swap.p2=0; tirage.t.swap.p2<instance.nb_procs; ++tirage.t.swap.p2) {
            unsigned int m2 = placement->assignment[tirage.t.swap.p2];
            if (m1!=m2
                && check_all_constraints_swap(placement, tirage.t.swap.p1, m1, tirage.t.swap.p2, m2)
                && ((gain=calcule_gain_tirage(initial, placement, tirage)) > gainMax))
            {
                best_p = tirage.t.swap.p2 ;
                gainMax = gain;
            }
        }
        //         DEBUG_PRINT ("temps=%.1f tirage_swap_checked: tirage.t.swap.p1 = %hu sur placement->assignment[tirage.t.swap.p1] = %hu et tirage.t.swap.p2 = %hu sur  placement->assignment[tirage.t.swap.p2] = %hu\n", t, tirage.t.swap.p1, placement->assignment[tirage.t.swap.p1], tirage.t.swap.p2, placement->assignment[tirage.t.swap.p2] );
    } 
    while ((t=tempsEcoule2(),tmax>t) && (best_p == NOPROCESS));

    tirage.t.swap.p2 = best_p;
    tirage.gain = gainMax;
    if (t>=tmax) tirage.nouveau = false;
    
    return tirage;
}



/**
 * couts
 */

int64 calcule_gain_tirage( const placement_t* initial, placement_t* courant, tirage_t tirage )
{
    switch ( tirage.type ) {
        case SHIFT:
            return calcule_gain_shift (initial, courant, tirage.t.shift.p, courant->assignment[tirage.t.shift.p], tirage.t.shift.m);
        case SWAP :
            return calcule_gain_swap(initial, courant, tirage.t.swap.p1, courant->assignment[tirage.t.swap.p1], tirage.t.swap.p2, courant->assignment[tirage.t.swap.p2]);
    }
    DEBUG_PRINT("Tirage aléatoire d'un mouvement de type inconnu\n");
    exit(-65);
}


/**
* déplacements
*/

void effectue_tirage( const placement_t * initial, placement_t * courant, tirage_t tirage )
{
    switch (tirage.type) {
        case SHIFT:
            deplace_proc(initial, courant, tirage.t.shift.p, tirage.t.shift.m);
            break;
        case SWAP:
            echange_procs(initial, courant, tirage.t.swap.p1, courant->assignment[tirage.t.swap.p1], tirage.t.swap.p2, courant->assignment[tirage.t.swap.p2] );
            break;
    }
}


#ifdef DEBUG
void DBG_AFFICHE_TIRAGE(tirage_t tirage, int64 cout)
{
    switch (tirage.type) {
        case SHIFT :
            DEBUG_PRINT("recuit: shift proc. %3d        vers %3d. Cout : %11lld\n", tirage.t.shift.p,  tirage.t.shift.m, cout);
            break;
        case SWAP :
            DEBUG_PRINT("recuit: swap proc. %3d           et %3d. Cout : %11lld\n", tirage.t.swap.p1, tirage.t.swap.p2, cout);
            break;
    }
}
#endif
