/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h> /* pour gettimeofday() */
#include <unistd.h>   /* pour getpid() */
#include <stdarg.h>   /* pour vararg */
#include <sys/resource.h>

#include <sys/sysinfo.h>
#include <strings.h>

#include "fmr.h"
#include "fmr_tools.h"

/* Initialisation du generateur de nombres aleatoires */

void init_rand( unsigned int seed, struct drand48_data * rand_buf) {
    struct timeval tv;
    
    gettimeofday( &tv, NULL );
    if (seed==0) {
       seed = ((unsigned) tv.tv_sec * 1000000u + (unsigned int) tv.tv_usec) << ( sizeof( unsigned int ) /2 );
       seed ^= (unsigned int) getpid();
    }
    
    DEBUG_PRINT("SEED = %u\n", seed);
    srand48_r( (signed) seed, rand_buf );
}



/* Delister un process d'ID passe en parametre.
 * Retourne un pointeur vers le process qu'on vient de delister.
 * La memoire de l'element qu'on deliste est liberee.
 * Retourne NULL si on n'a pas trouve le processus.
 */

processus_t* delister( struct process_list* liste, unsigned int id ) {

    struct process_list* enlisted_proc;
    struct process_list* previous_proc;
    processus_t* leproc;

    if( NULL == liste ) {
        return NULL;
    }

    enlisted_proc = liste;
    previous_proc = liste;

    do{
        if( enlisted_proc->process->id == id ) {
            previous_proc->next = enlisted_proc->next;
            leproc = enlisted_proc->process;
            free( enlisted_proc ); 
            return leproc;
        } else {
            previous_proc = enlisted_proc;
            enlisted_proc = enlisted_proc->next;
        }
    } while( enlisted_proc != NULL ) ;

    return NULL;
}

/* Enliste un processus a la fin d'une liste chainee.
 * Retourne la liste chainee.
 */

struct process_list* enlister( struct process_list* liste, processus_t* process  ) {

    struct process_list* current;
    
    if( NULL == liste ) {
        liste = (struct process_list*) malloc( sizeof( struct process_list ) );
        liste->next = NULL;
        liste->process = process;
        return liste;
    }

    current = liste;

    while( NULL != current->next  ) {
        current = current->next;
    }

    current->next = (struct process_list*) malloc( sizeof( struct process_list ) );
    current = current->next;
    current->next = NULL;
    current->process = process;

    return liste;
}

void
liste_init(liste_t *l, unsigned short nbelt) {
        l->prem_indice= LISTE_VIDE_US;
        l->nbMaxelt=nbelt;
	if (nbelt)
        	l->tab= (unsigned short *) malloc( nbelt*sizeof(unsigned short));
	else
		l->tab=NULL;
}

void
liste_free(liste_t *l) {
	free(l->tab);
}

/* on ajoute en tete de liste */
void
liste_in(unsigned short elt, liste_t *l) {
        if (l->prem_indice == LISTE_VIDE_US) {
                l->prem_indice = elt;
                l->tab[elt] = LISTE_VIDE_US;
        } else {
                l->tab[elt] = l->prem_indice;
                l->prem_indice = elt;
        }
}

/* ote l'elt elt de la liste */
unsigned short
liste_out(unsigned short elt, liste_t *l) {
        unsigned short i=l->prem_indice;

        if (i==elt) {
                l->prem_indice=l->tab[i];
                return(elt);
        }
        while (i!=LISTE_VIDE_US && l->tab[i]!=elt) {
                i=l->tab[i];
        }
        if (i== LISTE_VIDE_US)
                return(LISTE_VIDE_US); // pas trouvé
        l->tab[i] = l->tab[elt];
        l->tab[elt]= LISTE_VIDE_US;
        return(elt);
}


unsigned short
liste_alea(liste_t *l, struct drand48_data * rand_buf) {
	unsigned short i=l->prem_indice;
	unsigned short prev=i;
	double rval;
	while (i != LISTE_VIDE_US && (drand48_r(rand_buf, &rval),rval) >=0.5) {
		prev=i;
		i=l->tab[i];	
	}
	if (i== LISTE_VIDE_US)
	  return(prev);
	return(i);
}

/* copie de liste */
void
liste_dup(liste_t *dest, const liste_t *src, unsigned short nbelt) {
	liste_init(dest, nbelt);
	dest->prem_indice=src->prem_indice;
        memcpy(dest->tab, src->tab, sizeof(unsigned short)*nbelt);
        dest->nbMaxelt=nbelt;
}

/* copie de liste, suppose dest préalloué  à la bonne taille ! A utiliser avec attention.*/
void
liste_cpy(liste_t *dest, const liste_t *src, unsigned short nbelt) {
	if (dest==NULL || dest->nbMaxelt!=nbelt) {
		fprintf(stderr,"liste_cpy erreur:\n");
		exit(1);
	}
	dest->prem_indice=src->prem_indice;
        memcpy(dest->tab, src->tab, sizeof(unsigned short)*nbelt);
}

bool
liste_appartient(unsigned short elt, liste_t *l) {
        unsigned short i=l->prem_indice;

        while (i!=LISTE_VIDE_US && i!=elt) {
           i=l->tab[i];
        }
        if (i== LISTE_VIDE_US)
           return(false); // pas trouvé
        return(true);
}

void
liste_efface(liste_t *l) {
        l->prem_indice= LISTE_VIDE_US;
}

bool
liste_est_vide( liste_t *l) {
        return(l->prem_indice== LISTE_VIDE_US);
}

/* liste est réduite à un elt ? */
bool
liste_est_singleton( liste_t *l) {
        return(l->prem_indice != LISTE_VIDE_US && l->tab[l->prem_indice]== LISTE_VIDE_US);
}

void
liste_remplie( liste_t *l) {
	unsigned short i;

	l->prem_indice=0;
	for (i=0; i < l->nbMaxelt -1; i++) 
		l->tab[i]=(unsigned short)(i+1);
	l->tab[i]= LISTE_VIDE_US;
}

void
liste_fprint(FILE *fd, liste_t *l) {
        unsigned short i=l->prem_indice;

        while (i != LISTE_VIDE_US) {
                fprintf(fd, "%u,", i);
                i=l->tab[i];
        }
        fprintf(fd, "\n");
}

void
liste_print (liste_t *l) {
	liste_fprint(stdout, l);
}

unsigned short
liste_next(unsigned short e, liste_t *l) {
	return(l->tab[e]);
}

unsigned short
liste_head(liste_t *l) {
	return(l->prem_indice);
}

/************** traitement de listeCompacte_t ***************/
void
listeCompacte_alloc(listeCompacte_t *l, unsigned dimIndex, unsigned dimChainage) {
	l->tableauIndex = (unsigned short *) malloc (sizeof(unsigned short) * dimIndex);
	l->tableauChain = (unsigned short *) malloc (sizeof(unsigned short) * dimChainage);
}

void
listeCompacte_init(listeCompacte_t *l, unsigned dimIndex, unsigned dimChainage) {
	memset(l->tableauIndex, (int) LISTE_VIDE, sizeof(unsigned short) * dimIndex);
	memset(l->tableauChain, (int) LISTE_VIDE, sizeof(unsigned short) * dimChainage);
}

void
listeCompacte_in(listeCompacte_t *l, unsigned c, unsigned short id) {
	unsigned short i;
	if (l->tableauIndex[c] == LISTE_VIDE_US) {
		l->tableauIndex[c] =id;
	} else {
		i= l->tableauIndex[c];
		while (l->tableauChain[i] != LISTE_VIDE_US) {
			i= l->tableauChain[i];
		}
		l->tableauChain[i] =id;
	}
}

void
listeCompacte_out(listeCompacte_t *l, unsigned c, unsigned short id) {
	unsigned short i;
	if (l->tableauIndex[c] == id) {
		l->tableauIndex[c] = l->tableauChain[id];
	} else {
		i= l->tableauIndex[c];
		while (l->tableauChain[i] != id) {
			i= l->tableauChain[i];
			if (i== LISTE_VIDE_US ) {
				fprintf(stderr,"bug dans listeCompacte_out, %d non trouvé\n", id);
				exit(1);
			}
		}
		l->tableauChain[i]= l->tableauChain[id];
	}
	l->tableauChain[id] = LISTE_VIDE_US;
}

bool
listeCompacte_est_vide(listeCompacte_t *l, unsigned c) {
	return (l->tableauIndex[c] == LISTE_VIDE_US);
}

bool
listeCompacte_est_singleton(listeCompacte_t *l, unsigned c) {
	if (l->tableauIndex[c] == LISTE_VIDE_US)
		return(false);
	return (l->tableauChain[l->tableauIndex[c]] == LISTE_VIDE_US);
}

unsigned short
listeCompacte_tete(const listeCompacte_t *l, unsigned c) {
	return(l->tableauIndex[c]);
}

unsigned short
listeCompacte_next(const listeCompacte_t *l, unsigned c, unsigned short i) {
	return(l->tableauChain[i]);
}


unsigned short
listeCompacte_long(const listeCompacte_t *l, unsigned c) {
	unsigned short i,lalong=0;
	i= l->tableauIndex[c];
	while (i != LISTE_VIDE_US) {
		lalong++;
		i = l->tableauChain[i];
	}
	return(lalong);
}

unsigned short
listeCompacte_alea(const listeCompacte_t *l, unsigned c, struct drand48_data * rand_buf) {
	double rval;
	unsigned short i,ll=0; 
	unsigned short lalong= (unsigned short)(listeCompacte_long(l,c) * (drand48_r(rand_buf, &rval),rval));

	i=l->tableauIndex[c];
	while (i != LISTE_VIDE_US && ll<lalong) {
		ll++;
		i = l->tableauChain[i];
	}
	return i;
}

/* Suppose la structure preallouee */
void
listeCompacte_cpy (listeCompacte_t *dest, const listeCompacte_t const *src, unsigned dim, unsigned dimChainage) {
	memcpy (dest->tableauIndex, src->tableauIndex, sizeof(unsigned short) *dim);
	memcpy (dest->tableauChain, src->tableauChain, sizeof(unsigned short) *dimChainage);
}

void
listeCompacte_free (listeCompacte_t *l) {
	free(l->tableauIndex);
	free(l->tableauChain);
}

void
listeCompacte_fprint(FILE *f, listeCompacte_t *l, unsigned c) {
	unsigned short i;
	if (l->tableauIndex[c] == LISTE_VIDE_US) {
		fprintf(f,"Liste vide.");
	} else {
		i=l->tableauIndex[c];
		do {
			fprintf(f,"%hu,",i);
                        i= l->tableauChain[i];
                } while (i != LISTE_VIDE_US) ;
	}
}

/************** traitement de listeCompacte2D_t ***************/
void
listeCompacte2D_alloc(listeCompacte2D_t *l, unsigned int dimIndex1, unsigned int dimIndex2, unsigned int dimChainage) {
	l->tableauIndex = (unsigned short **) array2Dmalloc ( sizeof(unsigned short), dimIndex1, dimIndex2);
	l->tableauChain = (unsigned short *) malloc (sizeof(unsigned short) * dimChainage);
}

void
listeCompacte2D_init(listeCompacte2D_t *l, unsigned int dimIndex1, unsigned int dimIndex2, unsigned int dimChainage) {
	memset(l->tableauIndex[0], (int) LISTE_VIDE, sizeof(unsigned short) * dimIndex1*dimIndex2);
	memset(l->tableauChain, (int) LISTE_VIDE, sizeof(unsigned short) * dimChainage);
}

void
listeCompacte2D_in(listeCompacte2D_t *l, unsigned int c1, unsigned int c2, unsigned int id) {
	unsigned short i;
	if (l->tableauIndex[c1][c2] == LISTE_VIDE_US) {
		l->tableauIndex[c1][c2] = (unsigned short)id;
	} else {
		i= l->tableauIndex[c1][c2];
		while (l->tableauChain[i] != LISTE_VIDE_US) {
			i= l->tableauChain[i];
		}
		l->tableauChain[i] = (unsigned short)id;
	}
}

void
listeCompacte2D_out(listeCompacte2D_t *l, unsigned int c1, unsigned int c2, unsigned int id) {
	unsigned short i;
	if (l->tableauIndex[c1][c2] ==  (unsigned short)id) {
		l->tableauIndex[c1][c2] = l->tableauChain[id];
	} else {
		i= l->tableauIndex[c1][c2];
		while (l->tableauChain[i] !=  (unsigned short)id) {
			i= l->tableauChain[i];
			if (i==LISTE_VIDE_US ) {
				fprintf(stderr,"bug dans listeCompacte2D_out, %d non trouvé\n", id);
				exit(1);
			}
		}
		l->tableauChain[i]= l->tableauChain[id];
	}
	l->tableauChain[id] = LISTE_VIDE_US;
}

bool
listeCompacte2D_est_vide(listeCompacte2D_t *l, unsigned int c1, unsigned int c2) {
	return (l->tableauIndex[c1][c2] == LISTE_VIDE_US);
}

bool
listeCompacte2D_est_singleton(listeCompacte2D_t *l, unsigned int c1, unsigned int c2) {
	if (l->tableauIndex[c1][c2] == LISTE_VIDE_US)
		return(false);
	return (l->tableauChain[l->tableauIndex[c1][c2]] == LISTE_VIDE_US);
}

/* Suppose la structure preallouee */
void
listeCompacte2D_cpy (listeCompacte2D_t *dest, const listeCompacte2D_t const *src, unsigned int dim1, unsigned int dim2, unsigned int dimChainage) {
	array2Dcpy(dest->tableauIndex , src->tableauIndex, sizeof(unsigned short), dim1, dim2);
	memcpy (dest->tableauChain, src->tableauChain, sizeof(unsigned short) *dimChainage);
}

void
listeCompacte2D_free (listeCompacte2D_t *l) {
	free(l->tableauIndex);
	free(l->tableauChain);
}

void
listeCompacte2D_fprint(FILE *f, listeCompacte2D_t *l, unsigned int c1, unsigned int c2) {
	unsigned short i=l->tableauIndex[c1][c2] ;
	if (i == LISTE_VIDE_US)
		fprintf(f, "Liste vide.");
	else do {
		fprintf(f, "%hu,",i);
		i=l->tableauChain[i];
	} while (i!= LISTE_VIDE_US);
}


/************** traitement des array2D ******************/
void
array2DattribPtrs(void **ptr, size_t s, unsigned nb1D, unsigned int nb2D) {
   	unsigned i;
        int *p; /* les tailles de structures sont en multiples de sizeof(int) */
        if (ptr !=NULL) {
   		for (i=0, p=(int *)(ptr+nb1D); i< nb1D; i++, p= (void *)((char *)p +s*nb2D)) {
			ptr[i]= (void *)p;
		}
	}
}

void **
array2Dmalloc(size_t s, unsigned nb1D, unsigned nb2D) {
        void **ptr;

        ptr= (void **)malloc(s*nb1D*nb2D+ sizeof(void *)*nb1D);
	array2DattribPtrs(ptr, s, nb1D, nb2D);
	return(ptr);
}

	/* suppose que la zone a été allouée par array2Dmalloc */
void
array2Dmemset(void **ptr, int c, size_t s, unsigned nb1D, unsigned int nb2D) {
	memset(ptr[0], c, s*nb1D*nb2D);	
}

void **
array2Dcalloc(size_t s, unsigned nb1D, unsigned  nb2D) {
        void **ptr;

        ptr= (void **)malloc(s*nb1D*nb2D+ sizeof(void *)*nb1D);
        if (ptr !=NULL) {
	                /* division par sizeof(int) car le C va refaire cette multiplication
			 * à cause du type de p */
		array2DattribPtrs(ptr, s, nb1D, nb2D);
		memset(ptr[0], 0, s*nb1D*nb2D);
	}
	return(ptr);
}

void
array2Dcpy (void *dest, const void *src, size_t s, unsigned int nb1D, unsigned int nb2D) {
	array2DattribPtrs(dest, s, nb1D, nb2D);

	/* ne recopier que l'interieur de la matrice, conversion en void** pour
	 * que l'arithmetique de ptr fonctionne */
	memcpy ((void **)dest+nb1D, (void **)src+nb1D, s * nb1D * nb2D);
}

/** outils de gestion mémoire et temps */

/* retourne le nombre d'octets libre (sans compter le swap) */
unsigned long
memoireRestante() {
    struct sysinfo s_info;
    int error;

    error = sysinfo(&s_info);
    DEBUG_PRINT( "available memory: %lu / %lu (unit: %d)\n", s_info.freeram, s_info.totalram, s_info.mem_unit );

    return (s_info.freeram*s_info.mem_unit);
}

void
printMemoireUtilisee() {
	struct rusage ru;

	getrusage(RUSAGE_SELF, &ru);
        printf("getrusage: max_rss=%ld ko, unshared mem=%ld ko, pf=%ld\n", ru.ru_maxrss, ru.ru_idrss, ru.ru_majflt);
}


/* récupère le registre spécial compteur de tops d'horloge du proc INTEL */
#if defined(__i386__)
#warning "ATTENTION VOUS ETES EN 32 BITS"
static __inline__ unsigned long long rdtsc() {
  unsigned long long x;
  __asm__ volatile ("rdtsc" : "=A" (x));
  return x;
}
#elif defined(__x86_64__)
static __inline__ unsigned long long rdtsc() {
  unsigned long long a, d;
  __asm__ volatile ("rdtsc" : "=a" (a), "=d" (d));
  return (d<<32) | a;
}
#endif

void
tempsInit() {
#define MAXBUFLINE 200
	FILE *f;
	tempsZero = rdtsc();
	char buf[MAXBUFLINE];
	char *s;

	if ((f=fopen("/proc/cpuinfo","r"))==NULL) {
		fprintf(stderr,"impossible de lire /proc/cpuinfo !\n");
		exit(1);
	}
	do {
		s=fgets(buf, MAXBUFLINE, f);
	} while (strncmp(buf,"cpu MHz", strlen("cpu MHz"))!=0);
	sscanf(buf,"cpu MHz\t: %f\n", &cpufreq);
	DEBUG_STATS_PRINT("cpu freq=%f\n",cpufreq);
	/* Conversion en Hertz */
	cpufreq *= (float)1.0E6;
}


/* suppose que tempsinit() a été appelée avant, 
   suppose aussi que FREQCPU est correctement renseigné
   sur creos71 
   grep MHz /proc/cpuinfo 
   donne  cpu MHz         : 2533.224*/

/* Frequence du CPU pour creos71 en Hertz */
//#define FREQCPU 2533224000
/* Frequence CPU pour creos35 */
//#define FREQCPU 2392341000

float
tempsRestant(unsigned t) {
	DEBUG_STATS_PRINT("time_limit=%u\n",t);
	return ((float)t - (float)(rdtsc()-tempsZero)/(float)cpufreq);
}

float
tempsEcoule() {
	return ((float)(rdtsc()-tempsZero)/(float)cpufreq);
}

void 
tempsInit2() {
	gettimeofday(&tv_tempsInit, NULL);
}

double
tempsEcoule2() {
	struct timeval tv2;
	gettimeofday (&tv2, NULL);
	long long res= (tv2.tv_sec - tv_tempsInit.tv_sec)*1000000L + tv2.tv_usec - tv_tempsInit.tv_usec;

	return((double)res/1000000);
}

double
tempsRestant2() {
	return( (float)time_limit - tempsEcoule2());
}
